import numpy as np

from .fitter import Fitter
from .predictor import Predictor

from sklearn.utils.validation import check_X_y
from sklearn.utils.validation import check_array


class Mediator:
    """Mediator

    Class to check the input data, compute the parameters of the
    Sugeno classifier and do the classification.
    """

    def __init__(self):
        self.number_of_features = 0

    def check_train_data(self, X, y):
        """Check the input data for training.

        See 'check_X_y' in the documentation of scikit-learn for more
        information regarding the checks.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : arra-like of shape (1, n_samples)
            Target labels to X.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            Formated input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : arra-like of shape (1, n_samples)
            Formated target labels to X.

        Raises
        -------
        ValueError, if a check should fail.
        """

        X, y = check_X_y(X, y)

        if not self._check_for_regression_targets(y):
            raise ValueError('Unknown label type: ', y)

        return X, y

    def check_test_data(self, X):
        """Check the input data for classification.

        See 'check_array' in the documentation of scikit-learn for more
        information regarding the checks.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            Formated input data, where n_samples is the number of samples and
            n_features is the number of features.

        Raises
        -------
        ValueError, if a check should fail.
        """

        X = check_array(X)

        if np.shape(X)[1] != self.number_of_features:
            raise ValueError('Input data does not match number of features')

        return X

    def fit_components(self, X, y, maxitivity, margin, threshold):
        """Fit the Feature Transformation, the capacity function and
        the threshold for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like of shape (n_samples,)
            Target labels to X.

        maxitivity : int or None
            Maxitivity (First hyperparameter of the Sugeno classifier). If
            the maxitivity is None, the number of features will be used to
            compute the capacity function.

        margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        threshold : float or None
            Threshold. If None the linear program will be used to compute
            the threshold. Otherwise the linear program is skipped and the
            given threshold will be used for computing the capacity and
            the classification.

        Returns
        -------
        True, indicating that all parameters have been computed.
        """

        self.number_of_features = np.shape(X)[1]

        if maxitivity is not None and self.number_of_features < maxitivity:
            raise ValueError('Maxitivity is greater than n_features')

        fitter = Fitter()

        self.feature_transformation = fitter.fit_feature_transformation(X)

        normalized_X = self._get_normalized_X(X, self.feature_transformation)
        # print(normalized_X)

        if threshold is None:
            self.threshold = fitter.fit_threshold(normalized_X, y)
        else:
            self.threshold = threshold
        # print(self.threshold)

        self.capacity = fitter.fit_capacity(
            normalized_X, y, self.threshold, maxitivity, margin)
        # print(self.capacity)
        # self._print_capacity()

        return True

    def predict_classes(self, X):
        """Predict classes for input data.

        Predict classes for input data. The function fit_components
        had to be called before to use the Feature Transformation, the
        capacity function and the threshold.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        result : ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        predictor = Predictor()

        normalized_X = self._get_normalized_X(X, self.feature_transformation)

        result = predictor.get_classes(
            normalized_X, self.capacity, self.threshold)

        return result.ravel()

    def _check_for_regression_targets(self, y):
        """Check target data for regression targets.

        Regression targets are considered to be real non integer numbers.
        This check is necessary to be compatible with scikit-learn.

        Parameters
        -------
        y : array-like of shape (1,)
            Target labels.

        Returns
        -------
        False, if there is a regression target. Otherwise True.
        """

        for value in y:
            # check for numeric value
            if not (isinstance(value, (int, float))
               and not isinstance(value, bool)):
                continue

            if not float(value).is_integer():
                return False

        return True

    def _get_normalized_X(self, X, f):
        """Normalize the input data using a Feature Transformation.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        f : FeatureTransformation
            Feature Transformation.

        Returns
        -------
        result : ndarray of shape (n_samples, n_features)
            Normalized input data, where n_samples is the number of samples
            and n_features the number of features.
        """

        result = list()

        for x in X:
            normalized_x = self._get_normalized_x(x, f)

            result.append(normalized_x)

        return np.array(result)

    def _get_normalized_x(self, x, f):
        """Normalize example using a Feature Transformation.

        Parameters
        -------
        x : array-like of shape (1, n_features)
            Example, where n_features is the number of features.

        f : FeatureTransformation
            Feature Transformation.

        Returns
        -------
        normalized_x : ndarray
            Normalized example.
        """

        normalized_x = list()
        number_of_features = len(x)

        for i in range(number_of_features):
            feature = x[i]
            normalized_feature = f[i](feature)

            normalized_x.append(normalized_feature)

        return np.array(normalized_x)

    def _print_capacity(self):
        for key, value in self.capacity.items():
            print(key, '->', value)



import numpy as np

from .components.feature_transformation import FeatureTransformation
from .components.capacity import Capacity
from .components.threshold import Threshold


class Fitter:
    """Fitter

    This class implements the funtions to compute the parameters of the
    Sugeno classifier.
    """

    def __init__(self):
        pass

    def fit_feature_transformation(self, X):
        """Fit the Feature Transformation for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        feature_transformation : FeatureTransformation
            Instance of the fitted Feature Transformation.
        """

        feature_transformation = FeatureTransformation(X)

        return feature_transformation

    def fit_capacity(self, X, y, threshold, maxitivity, margin):
        """Fit the capacity function for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : arra-like of shape (1, n_samples)
            Target labels to X.

        threshold : float
            Threshold.

        maxitivity : int or None
            Maxitivity (First hyperparameter of the Sugeno classifier). If
            the maxitivity is None, the number of features will be used to
            compute the capacity function.

        margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        Returns
        -------
        function : dict
            Dictionary describing the capacity function.
        """

        number_of_features = np.shape(X)[1]

        if maxitivity is None:
            maxitivity = number_of_features

        capacity = Capacity(X, y)

        function = capacity.compute_capacity(threshold, maxitivity, margin)

        return function

    def fit_threshold(self, X, y):
        """Fit the threshold for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : arra-like of shape (1, n_samples)
            Target labels to X.

        Returns
        -------
        threshold : float
            Threshold.
        """

        threshold = Threshold(X, y)

        result = threshold.compute_threshold()

        return result



import numpy as np
import pulp as pl

from pulp import LpProblem
from pulp import LpMinimize
from pulp import LpVariable
from pulp import lpSum


class Threshold:
    """Threshold class to compute the value for the threshold

    Use the constructor to store the train data and the function
    compute_threshold to compute the threshold. The train data are
    supposed to be normalized.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Input data, where n_samples is the number of samples and
        n_features is the number of features. The data are supposed
        to be normalized. Otherwise the linear programm might not
        have a solution.

    y : array-like (1, n_samples)
        Target labels to X.
    """

    def __init__(self, X, y):
        self.X = X
        self.y = y

    def compute_threshold(self):
        """Compute the threshold according to the definition of the
        linear program.

        Returns
        -------
        threshold : float
            Value for the threshold.
        """
		
        threshold = self._solve_linear_program()

        return threshold

    def _solve_linear_program(self):
        """Solve linear program defined for the threshold.

        Returns
        -------
        threshold : float
            Value for the threshold.
        """

        # for each instance in the train data a contraint is derived
        number_of_constraints = np.shape(self.X)[0]

        # --------------------- Initialize --------------------- #
        # interior point: , options=['barrier']
        solver = pl.PULP_CBC_CMD(msg=0)  # , options=['barr'])

        linprob = LpProblem(name='Threshold', sense=LpMinimize)

        beta = LpVariable(name='beta', lowBound=0, upBound=1, cat='Continuous')

        slack_items = np.arange(number_of_constraints)
        slack_variables = LpVariable.dicts(
            's', slack_items, lowBound=0, upBound=1, cat='Continuous')

        linprob += lpSum([1 * slack_variables[i]
                          for i in range(number_of_constraints)])

        # --------------------- Initialize --------------------- #

        # --------------------- Constraints --------------------- #

        for i in range(number_of_constraints):
            x = self.X[i]
            y = self.y[i]

            median = np.median(x)

            if y == 1:
                linprob += (median + slack_variables[i] >= beta)
            else:
                linprob += (median - slack_variables[i] <= beta)

        # --------------------- Constraints --------------------- #
        linprob.solve(solver)

        # search for correct variable
        for variable in linprob.variables():
            if str(variable).startswith('beta'):
                return variable.varValue

        raise ValueError('Unable to solve linear program for threshold')



import numpy as np

from . import helper as h


class SugenoIntegral:
    """Sugeno integral

    Class to store a capacity function and compute the aggregation value.

    Parameters
    -------
    capacity : dict
            Dictionary describing the capacity function.
    """

    def __init__(self, capacity):
        self.capacity = capacity

    def compute_aggregation_value(self, x, method='definition'):
        """Compute the aggregation value for a given example.

        Parameters
        -------
        x : array-like of shape (1, n_features)
                Example, where n_features is the number of features.

        method : str
                Specify the method used for the Sugeno integral. Default is
                the method according to the definition. Use 'median' to use
                the median representation of the Sugeno integral.

        Returns
        -------
        integral_value : float
                Integral value computed by the specified method.

        """

        if method == 'definition':
            integral_value = self._get_integral_value(x, self.capacity)

            return integral_value

        if method == 'median':
            integral_value = self._get_median_value(x, self.capacity)

            return integral_value

    def _get_integral_value(self, x, capacity):
        """Get integral value for a given example according the definition
        of the Sugeno integral.

        Parameters
        -------
        x : array-like of (1, n_features)
                Example, where n_features is the number of features.

        capacity : dict
                Dictionary describing the capacity function.

        Returns
        -------
        result : float
                Integral value.
        """

        result = 0
        number_of_features = len(x)
        sorted_x = np.sort(x)

        for j in range(number_of_features):
            x_value = sorted_x[j]

            feature_subset = h.get_feature_subset(x, j + 1)
            subset_value = capacity[feature_subset]

            minimum_for_index = min(x_value, subset_value)

            # update maximum
            if minimum_for_index > result:
                result = minimum_for_index

        return result

    def _get_median_value(self, x, capacity):
        """Get integral value for a given example according the median
        representation of the Sugeno integral.

        Parameters
        -------
        x : array-like of (1, n_features)
                Example, where n_features is the number of features.

        capacity : dict
                Dictionary describing the capacity function.

        Returns
        -------
        result : float
                Integral value.
        """

        sorted_x = np.sort(x)

        median_representation = list(x)

        for i in range(2, len(x) + 1):
            subset_for_index = h.get_feature_subset(x, i)

            subset_value = capacity[subset_for_index]

            median_representation.append(subset_value)

        integral_value = np.median(median_representation)

        return integral_value



import numpy as np

from .components.sugeno_integral import SugenoIntegral


class Predictor:
    """Predictor

    This class implements the methods for the classification.
    """

    def __init__(self):
        pass

    def get_classes(self, X, capacity, threshold):
        """Get classes for input data.

        Get classes for input data. Use the capacity for the Sugeno integral
        and the threshold for the classification.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        capacity : dict
            Dictionary describing the capacity function.

        threshold : float
            Threshold.

        Returns
        -------
        result : ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        sugeno_integral = SugenoIntegral(capacity)

        result = self._get_classes_for_X(X, sugeno_integral, threshold)

        return result

    def _get_classes_for_X(self, X, integral, threshold):
        """Get classes for input data.

        Get classes for input data. Use the integral and the threshold
        for the classification.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        integral : SugenoIntegral
            Sugeno integral.

        threshold : float
            Threshold.

        Returns
        -------
        result : ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        result = list()

        for x in X:
            # aggregation METHOD , method='median'
            aggregation_value = integral.compute_aggregation_value(x)

            cl = self._get_decision(aggregation_value, threshold)

            result.append(cl)

        return np.array(result)

    def _get_decision(self, aggregation_value, threshold):
        """Get class for aggregation value.

        Parameters
        -------
        aggregation_value : float
            Aggretion value.

        threshold : float
            Threshold.
        """

        if aggregation_value >= threshold:
            return np.array([1])
        else:
            return np.array([0])



import numpy as np


def get_feature_subset(x, index):
    """Get the feature subset A_{s(index)} for a given example and a
    permutation s.

    Parameters
    -------
    x : array-like of shape (1, n_features)
        Example, where n_features is the number of features.

    index : int
        Index for the feature subset.

    Returns
    -------
    result : frozenset
        Feature subset.
    """

    result = set()

    m = len(x)
    sorted_x = np.sort(x)

    permutation = _get_permutation_for_x(x, sorted_x)

    # compute feature subset {\sigma(i),...,\sigma(m)}
    for i in range(index, m + 1):
        result.add(permutation[i])

    return frozenset(result)


def _get_permutation_for_x(x, sorted_x):
    """Get permutation for an example.

    Get a dictionary describing a permutation s of x with
    x_s(1) <= ... <= x_s(m). The dictionary maps the values
    1,...,m to the corresponding position of the permutation,
    where m is the number of features.

    Parameters
    -------
    x : array-like of shape (1, n_features)
        Example, where n_features is the number of features.

    sorted_x : array-like of shape (1, n_features)
        Sorted example, where n_features is the number of features.

    Returns
    -------
    permutation : dict
        Dictionary describing the permutation.
    """

    permutation = dict()

    x_copy = x.copy()

    for i in range(len(x)):
        value = sorted_x[i]

        # TODO which index to choose if there is a tie between values
        index = np.where(x_copy == value)[0][0]

        permutation[i + 1] = index + 1
        x_copy[index] = -9999  # TODO choose dummy value

    return permutation



import numpy as np


class FeatureTransformation:
    """Feature Transformation

    Funtion f(x)=(f_0(x_0),...,f_{m-1}(x_{m-1})) to normalize the train
    data, where m is the number of features. Use the item and call
    operator to get the value for a specific x_i.
    Example: f[1](5) is equivalent to f_1(5) for x_i=5

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Input data, where n_samples is the number of samples and
        n_features is the number of features.
    """

    def __init__(self, X):
        self.function = self._initialize(X)

    def _initialize(self, X):
        """Initialize the funtion.

        Create an array of shape (1, n_features), where n_features is
        the number of features. Each index stores an instance of the
        class _FeatureTransformationComponent describing a component f_i

        Parameters:
        ----------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns:
        ----------
        array : ndarray of shape (, n_features)
            Array containing a _FeatureTransformationCompoent f_i for
            each feature.
        """

        result = list()

        number_of_columns = np.shape(X)[1]

        for i in range(number_of_columns):
            # create a feature transformation component for eacht feature
            column = X[:, i]

            function_component = _FeatureTransformationComponent(column)

            result.append(function_component)

        return np.array(result)

    def __getitem__(self, i):
        # use the item operator to access the feature transformation
        # compontent f_i
        return self.function[i]


class _FeatureTransformationComponent:
    """Feature Transformation Component

    A feature transformation component f_i. Store all values of the i-th
    feature. Use the call operator to compute the value for a specific x_i.

    Parameters:
    ----------
    x_values : array-like of shape (n_samples, 1)
        Input data for the i-th feature, where n_samples is the number
        of samples
    """

    def __init__(self, x_values):
        self.x_values = np.sort(x_values)

        # corresponding y values according to a_j
        self.y_values = np.array(
            [1/2 * (
                self.x_values[self.x_values < value].size
                + self.x_values[self.x_values <= value].size
            ) for value in self.x_values]
        )

        # normalize y-values
        self.y_values /= self.x_values.size

        # make all values unique
        self.x_values = np.unique(self.x_values)
        self.y_values = np.unique(self.y_values)

        assert(len(self.x_values) == len(self.y_values))

    def __call__(self, x):
        # use the call operator to compute a normalized value for x
        return np.interp(x, self.x_values, self.y_values, left=0, right=1)



import numpy as np
import pulp as pl
import itertools

from . import helper as h

from collections import defaultdict
from pulp import LpProblem
from pulp import LpMinimize
from pulp import LpVariable
from pulp import lpSum


class Capacity:
    """Capacity function for the sugeno integral

    Normalized Capacity funtion for the sugeno integral. Use the function
    compute_capacity to compute the capacity. The train data are supposed
    to be normalized.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Input data, where n_samples is the number of samples and
        n_features is the number of features. The data are supposed
        to be normalized. Otherwise thelinear programm might not have
        a solution.

    y : array-like (1, n_samples)
        Target labels to X.
    """

    def __init__(self, X, y):
        self.X = X
        self.y = y

        self.number_of_features = np.shape(self.X)[1]

        # store the values 1...m, i.e. [m]
        self.base_set = frozenset(
            [i for i in range(1, self.number_of_features + 1)])

        # data structures, remove empty set because it is set to 0
        self.powerset = self._get_powerset(self.base_set)[1:]
        self.subset_ids = self._get_subset_ids(self.powerset)
        self.subsets_of_size = self._get_subsets_of_size(self.powerset)

    def compute_capacity(self, threshold, maxitivity, margin):
        """Get the capacity function for the saved input data.

        Parameters
        ----------
        threshold : float
            Threshold.

        maxitivity : int
            Maxitivity (First hyperparameter of the sugeno classifier).

        margin : float
            Margin (Second hyperparameter of the sugeno classifier).

        Returns
        ----------
        function : dict
            Capacity function: the keys are the subsets of [m], the
            values are the computed values.
        """

        linprob, number_of_c_variables = self._initialize_linear_program(
            threshold, maxitivity, margin)

        result = self._solve_linear_program(linprob, number_of_c_variables)

        function = self._get_capacity_from_result(result, maxitivity)

        function = self._update_capacity_for_maxitivity(maxitivity, function)

        return function

    def _get_powerset(self, s):
        """Get the powerset of s.

        Parameters
        ----------
        s : array-like of shape (1, n_samples)
            Input data, where n_samples is the number of elements in the
            set.

        Returns
        ----------
        powerset : ndarray of shape (1, n_samples)
            Array containing all subsets of given set.
        """

        result = list()
        items = list(s)

        # copied from
        # https://docs.python.org/3/library/itertools.html#itertools-recipes
        powerset = itertools.chain.from_iterable(
            itertools.combinations(items, r) for r in range(len(items) + 1))

        # parse all elements to frozensets
        for item in powerset:
            result.append(frozenset(item))

        return np.array(result)

    # map: subset -> id
    def _get_subset_ids(self, s):
        """Get a map: subset -> id

        Compute a dictionary describing a map. The keys are subsets of s,
        which are mapped to an unique id ranging from 0 to |s|-1

        Parameter
        ----------
        s : array-like of shape (1,)
            Sets, which are supposed to be mapped.

        Returns
        ----------
        map : dict
            Dictionary, which maps the given subsets to an unique id
        """

        result = dict()
        counter = 0

        for subset in s:
            result[subset] = counter
            counter += 1

        return result

    # map: n -> list with subsets of size n
    def _get_subsets_of_size(self, s):
        """Get a map: n -> list with subsets of size n

        Compute a dictionaty describing a map. The keys are numerical
        values ranging from the smallest element to the biggest element
        in s. These are mapped to a list containing all elements of the
        same size as the key.

        Parameter
        ----------
        s : array-like of shape (1, n_sets)
            Input. Each element is a set of type frozenset.

        Returns
        ----------
        map : dict
            Dictionary, which maps the numerical values to a list
            containing all sets with the corresponding value size.
        """

        result = defaultdict(list)

        for subset in s:
            result[len(subset)].append(subset)

        return result

    def _get_capacity_from_result(self, result, maxitivity):
        """Get a dictionary containing all computed values for given variables

        Compute first part of the capacity funtion. This method sets all
        values for the sets, which corresponding variables were part of
        the linear program.

        Parameter
        ----------
        result : list
            List with values for the subsets. The position in the list
            has to match the index of the corresponding set.

        maxitivity : int
            The maxitivity used by solving the linear program.

        Returns
        ----------
        function : dict
            Dictionary containing the sets as keys and their
            corresponding computed values as values.

        """

        function = dict()

        subsets_of_size_k = self.subsets_of_size[maxitivity]

        for s in self.powerset:
            if len(s) <= maxitivity:
                subset_id = self.subset_ids[s]

                # set values from variables of LP
                result_value = result[subset_id]

                function[s] = result_value

        # base cases
        function[frozenset()] = 0.0
        function[self.base_set] = 1.0

        return function

    def _update_capacity_for_maxitivity(self, maxitivity, capacity):
        """Update a given capacity function for maxitivity k.

        If the maxitivity k is smaller than the number of features,
        the values for sets of size s with s > k are not part of the
        linear program. Hence they have to be added manually.

        Parameters
        ----------
        maxitivity : int
            Maxitivity of the capacity function.

        capacity : dict
            Dictionary describing the capacity funtion. All values for
            sets smaller or equal to the maxitivity have to be set.

        Returns
        ----------
        capacity : dict
            Updated dictionary containing all subsets of the base set.
        """

        subsets_of_size_k = self.subsets_of_size[maxitivity]

        # set values for all sets with size greater than the maxitivity
        for s in range(maxitivity + 1, self.number_of_features):
            subsets_of_size_s = self.subsets_of_size[s]

            for subset in subsets_of_size_s:
                # get all subsets of size k for the iterated set
                k_subsets = self._get_k_subsets_for_set(maxitivity, subset)

                # get maximum value computed for the capacity
                maximum = self._get_maximum_of_subsets(k_subsets, capacity)

                capacity[subset] = maximum

        return capacity

    def _get_k_subsets_for_set(self, k, subset):
        """Get all subsets of size k for a given set.

        Parameters
        ----------
        k : int
            Size of the sets to be determined.

        subset : frozenset
            Set whose subsets of size k to be determined.

        Returns
        ----------
        result : ndarray
            Array containing all subsets of size k for subset.
        """

        result = list()

        k_subsets = self.subsets_of_size[k]

        # find all subsets of size k
        for s in k_subsets:
            if s.issubset(subset):
                result.append(s)

        return np.array(result)

    def _get_maximum_of_subsets(self, subsets, function):
        """Find the maximum value for the subsets in the given function.

        Parameters
        ----------
        subsets : array-like of shape (1,)
            Subsets whose maximum to be determined.

        function : dict
            Dictionary mapping the susbets to a value.

        Returns
        ----------
        result : float
            Maximum of function values for the given subsets.
        """

        result = 0

        # find maximum of subsets
        for s in subsets:
            if function[s] > result:
                result = function[s]

        return result

    def _solve_linear_program(self, linprob, number_of_c_variables):
        """Solve the linear program.

        Parameters
        ----------
        linprob : LpProblem
            Instance of a linear program to be solved.

        number_of_c_variables : int
            Number of variables for the function values. This is used because
            not all c_variables might be part of the linear program.

        Returns
        ----------
        solution : ndarray
            Array containing the solution of the x-variables from the
            linear program. The positions in the array match the ids
            from subset_ids.
        """

        # interior point: , options=['barr']
        solver = pl.PULP_CBC_CMD(msg=0)

        linprob.solve(solver)

        solution = np.zeros(number_of_c_variables)

        # find x-variables and set their values
        for variable in linprob.variables():
            if str(variable).startswith('x'):
                index = str(variable)[len('x') + 1:]

                solution[int(index)] = variable.varValue

        return solution

    def _initialize_linear_program(self, threshold, maxitivity, margin):
        """Initialize the linear program.

        Intialize the linear program, i.e. the variables, the objective
        function and the contraints. The variables are named x_0,...,x_s
        depending on the number of features and the maxitivity.

        Parameters
        ----------
        threshold : float
            Threshold.

        maxitivity : int
            Maxitivity (First hyperparameter of the Sugeno classifier).

        margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        Returns
        ----------
        linprob : LpProblem
            Instance of the linear problem.

        number_of_c_variables : int
            Number of variables to compute for the capacity function values.
        """

        number_of_c_variables = self._get_number_of_values_to_compute(
            1, maxitivity)
        number_of_instances = np.shape(self.X)[0]

        # --------------------- Initialize --------------------- #

        linprob = LpProblem('Capacity', sense=LpMinimize)

        c_items = np.arange(number_of_c_variables)
        c_variables = LpVariable.dicts(
            'x', c_items, lowBound=0, upBound=1, cat='Continuous')

        # --------------------- Initialize --------------------- #

        # --------------------- monotony constraints --------------------- #

        monotony_constraints = self._get_monotony_constraints(1, maxitivity)

        for A, B in monotony_constraints:
            A_id = self.subset_ids[A]
            B_id = self.subset_ids[B]

            linprob += (c_variables[A_id] <= c_variables[B_id])

        # --------------------- monotony constraints --------------------- #

        # --------------------- median constraints --------------------- #

        slack_item = 0
        slack_variables = dict()

        for i in range(number_of_instances):
            x = self.X[i]
            y = self.y[i]

            # get all sets which have to be constrained
            feature_subsets = self._get_feature_subsets_for_x_y(
                x, y, threshold, maxitivity, margin)

            for feature_subset in feature_subsets:
                assert len(feature_subset) <= maxitivity

                # create new slack variable
                s_variable = LpVariable(
                    's_' + str(slack_item),
                    lowBound=0,
                    upBound=1,
                    cat='Continuous'
                )
                slack_variables[slack_item] = s_variable

                feature_subset_id = self.subset_ids[feature_subset]

                if y == 1:
                    linprob += (c_variables[feature_subset_id]
                                + slack_variables[slack_item]
                                >= threshold + margin)
                else:
                    linprob += (c_variables[feature_subset_id]
                                - slack_variables[slack_item]
                                <= threshold - margin)

                slack_item += 1

        number_of_s_variables = slack_item

        # --------------------- median constraints --------------------- #

        # objective function
        linprob += lpSum([1 * slack_variables[i]
                          for i in range(number_of_s_variables)])

        # return linprob
        return linprob, number_of_c_variables

    def _get_number_of_values_to_compute(self, lower, upper):
        """Get number of values to compute.

        Get the number of variables which correspond to the C variables.
        This number is equivalent to the sum of the binomial coefficients
        <m over lower> to <m over upper>, where m is the nubmer of features.

        Parameter
        ----------
        lower : int
             Lower bound.

        upper : int
            Upper bound.

        Returns
        ----------
        result : int
            Number of needed C variables.

        """

        result = 0

        for i in range(lower, upper + 1):
            subsets_of_size_i = self.subsets_of_size[i]

            result += len(subsets_of_size_i)

        return result

    # create monotony constraints (equation (7))
    def _get_monotony_constraints(self, lower, upper):
        """Get monotony contraints for the linear program.

        Get all contraints for f(A) <= f(B) for the capacity funtion f
        two sets A and B with |B|=|A| + 1 and A is a subet of B.

        Parameters
        ----------
        lower : int
            Lower bound describing the minimum size of sets.

        upper : int
            Upper bound describung the maximum size of sets

        Returns
        ----------
        constraints : ndarray of shape (1,)
            Array containing tuples (A,B) corresponding to sets A and B
            according to the description above.

        """

        monotony_constraints = list()

        for s in range(lower, upper):
            # get contraints for f(A) <= f(B) with |A| = s
            constraints = self._get_constraints_for_subsets_of_size(s)

            for item in constraints:
                monotony_constraints.append(item)

        return np.array(monotony_constraints)

    # contraints: subset A of size i <= subset B of size i+1
    def _get_constraints_for_subsets_of_size(self, size):
        """Get constraints for all subsets of size.

        Get all contraints f(A) <= f(B) with |A| = size and
        |B| = size + 1.

        Parameters
        ----------
        size : int
            Size of subsets.

        Returns
        ----------
        contraints : ndarray of shape (1,)
            Array containing tuples (A,B) corresponding to sets A and B
            according to the description above.
        """

        constraints = list()

        # check pairwise subset relation of sets of size and sets of size + 1
        for subset in self.subsets_of_size[size]:
            for other_subset in self.subsets_of_size[size + 1]:

                if subset.issubset(other_subset):
                    constraints.append((subset, other_subset))

        return np.array(constraints)

    def _get_feature_subsets_for_x_y(self, x, y, threshold, maxitivity,
                                     margin):
        """Get sets which have to be constraint for a given example.

        Parameters
        ----------
        x : array-like of shape (1, n_features)
            Example, where n_feature is the number of features.

        y : int
            Target label to x.

        threshold : float
            Treshold.

        maxitivity : int
            Maxitivity (First hyperparameter of the Sugeno classifier).

        margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        Returns
        ----------
        result : ndarray of shape (1,)
            Array containing all sets, which have to be constraint.
        """

        result = list()

        sorted_x = np.sort(x)

        if y == 1:
            p = self._get_p_for_positive_x(
                sorted_x, threshold, maxitivity, margin)
        else:
            p = self._get_p_for_negative_x(
                sorted_x, threshold, maxitivity, margin)

        # no contraint has to be added
        if p is None:
            return np.array([])

        if y == 1:
            if p > maxitivity:
                # to be consistent always choose c_A with size k
                feature_subset = h.get_feature_subset(
                    x, self.number_of_features - maxitivity + 1)
            else:
                feature_subset = h.get_feature_subset(
                    x, self.number_of_features - p + 1)

            result.append(feature_subset)
        else:
            if p < self.number_of_features - maxitivity:
                # add all subsets of B with size k
                B = h.get_feature_subset(x, p + 1)

                feature_subsets = self._get_feature_subsets_of_size(
                    maxitivity, B)

                for feature_subset in feature_subsets:
                    result.append(feature_subset)
            else:
                feature_subset = h.get_feature_subset(x, p + 1)

                result.append(feature_subset)

        return np.array(result)

    def _get_p_for_positive_x(self, x, threshold, maxitivity, margin):
        """Get the p value for a positive example (y = 1).

        Parameters
        ----------
        x : array-like of shape (1, n_features)
            Example, where n_features is the number of features.

        threshold : float
            Threshold.

        maxitivity : int
            Maxitivity (First hyperparameter of the Sugeno classifier).

        Margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        Returns
        ----------
        p : int or None
            p value describing the set which has to be contraint or None
            if no constraint can be derived.
        """

        # number of elements >= threshold + margin
        p = len(np.where(x >= threshold + margin)[0])

        if p == self.number_of_features or p == 0:
            return None

        return p

    def _get_p_for_negative_x(self, x, threshold, maxitivity, margin):
        """Get the p value for a negative example (y = 0).

        Parameters
        ----------
        x : array-like of shape (1, n_features)
            Example, where n_features is the number of features.

        threshold : float
            Threshold.

        maxitivity : int
            Maxitivity (First hyperparameter of the Sugeno classifier).

        Margin : float
            Margin (Second hyperparameter of the Sugeno classifier).

        Returns
        ----------
        p : int or None
            p value describing the set which has to be contraint or None
            if no constraint can be derived.
        """

        # TODO less or less equal ?
        # number of elments < threshold - margin
        p = len(np.where(x < threshold - margin)[0])

        if p == self.number_of_features or p == 0:
            return None

        return p

    # get all subsets A of size k with A<B={p,...,m}, s.t. c(A) < threshold
    def _get_feature_subsets_of_size(self, k, B):
        """Get feature subsets of size k.

        Get all feature subsets of size k for the set B.

        Parameters
        ----------
        k : int
            Size of feature subsets to be determined.

        B : frozenset
            Set whose feature subsets to be determined.

        Returns
        ----------
        result : ndarray of shape (1,)
            Array containting all feature subsets which have to be constraint.
        """

        result = list()

        # there are only subsets with less size
        if len(B) < k:
            return np.array(result)

        for A in self.subsets_of_size[k]:
            if A.issubset(B):
                result.append(A)

        return np.array(result)

    def print_capacity(self, function):
        """Print capacity funtion.

        Parameters
        ----------
        funtion : dict
            Dictionary describing the capacity funtion.

        Returns
        ----------
        result : str
            String describing the capacity function in the format
            'set -> value'.
        """

        result = ''

        for key, value in function.items():
            result += str(key) + ' -> ' + str(value) + '\n'

        return result



import numpy as np

from .mediator.mediator import Mediator

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import check_X_y
from sklearn.utils.validation import check_array


class SugenoClassifier(BaseEstimator, ClassifierMixin):
    """The Sugeno classifier.

    Implementation of the Sugeno classifier, which were invented in
    the paper "Machine Learning with the Sugeno Integral: The Case of
    Binary Classification". The classifier is compatible to scikit-learn,
    i.e. it can be used together with other algorithms from this library.

    Parameters
    -------
    maxitivity : int, default=None
        The maxitivity of the capacity function. Setting this parameter
        might increase the performance. The value is expected to be
        between 1 and the number of features in a dataset. The default
        value will set the maxitivity to the number of features in a
        given dataset.

    margin : float, default=0
        The margin which influences the values of the capacity function.
        Tests have shown that a better performance can be achieved by
        choosing a margin greater than 0. Reasonable values can be found
        in the intervals [0, 0.1] or [0, 0.2].

    threshold : float, default=None
        The threshold which is used to compute the capacity and classify
        data. Setting this parameter skips the computation of the liear
        program and uses the value instead. This parameter was used in
        an evaluation part of a Bachelor thesis and should not be changed.
    """

    def __init__(self, maxitivity=None, margin=0, threshold=None):
        self.maxitivity = maxitivity
        self.margin = margin
        self.threshold = threshold

    def fit(self, X, y):
        """Initialize the parameters of the Sugeno classifier.

        Initialize the Feature Transformation, the capacity and the
        threshold with the initializes hyperparameter for a given dataset.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features. The numger of features
            have to less or equal to the maxitivity.

        y : array-like of shape (n_samples,)
            Target labels to X.

        Returns
        -------
        self : SugenoClassifier
            Fitted estimator.
        """

        self.mediator_ = Mediator()

        X, y = self.mediator_.check_train_data(X, y)

        # replace class labels with the values 0 and 1
        self.classes_, y = np.unique(y, return_inverse=True)

        self.mediator_.fit_components(
            X, y, self.maxitivity, self.margin, self.threshold)

        # attribute which is used to be compatible with scikit-learn
        self.n_features_in_ = self.mediator_.number_of_features

        return self

    def predict(self, X):
        """Predict class for X.

        Predict the class labels for all samples in X, which were
        specified in fit. The number of features have to match the
        the number of features from the train data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted classes.
        """

        check_is_fitted(self)

        X = self.mediator_.check_test_data(X)

        result = self.mediator_.predict_classes(X)

        return self.classes_[result]

    # ===============================================================
    # Functions for compatibility with scikit-learn. They are not
    # supposed to be used.
    # ===============================================================

    def get_params(self, deep=True):
        return {'maxitivity': self.maxitivity,
                'margin': self.margin,
                'threshold': self.threshold}

    def _more_tags(self):
        return {'binary_only': True, 'poor_score': True}

    def _get_threshold(self):
        return self.mediator_.threshold
